<?php
namespace com\dunweber\searchengine;
// ========================================================================
// PHP Search Engine
// Version: 2019-10-15 21:37 (UTC+02:00)
// https://www.dunweber.com/docs/scripts/releases/searchengine
// Copyright (C) Christian L. Dünweber
// This program is licensed under the MIT License:
//    https://opensource.org/licenses/MIT
// ------------------------------------------------------------------------
// Usage:
// require_once('SearchEngine.php');
// $se = new com\dunweber\searchengine\SearchEngine('.'); // Set start dir
// echo $se->getSearchModule();
// ------------------------------------------------------------------------
// TODO: If you page titles are stored in PHP variables modify getTitle().
// ========================================================================

class Conf
{
	public const

	// File types to open and search inside. Others will only participate with filename
	SEARCH_SOURCE_TYPES = ['c','h','m','inc','java','js','php','phps','sql','htm','html','css','xml','txt','csv','bat','cmd','ps1','json',
		'doc','docx','xls','xlsx','ppt','pptx','pdf',],

	// The maximum file size in bytes to open and search in
	FILESIZE_MAX_SEARCH_OPEN = 1024*1024; // 1 MB
}

// ========================================================================
// Implementation
// ------------------------------------------------------------------------
class SearchEngine
{
	private $dir; // Starting dir
	private $allFiles = []; // All files read in current dir

	// ====================================================================
	// Constructor.
	// --------------------------------------------------------------------
	public function __construct(string $dir)
	{
		$this->dir = $dir;
	}

	// ====================================================================
	// Get the search module.
	// --------------------------------------------------------------------
	public function getSearchModule(): string
	{
		$dir = realpath($this->dir); // Get current abs dir
		$webdir = $this->getWebPath($dir); // Get current web dir 
		$q = trim($_GET['q'] ?? ''); // Get search string
		$wholeWord = isset($_GET['wholeWord']); // Has no effect if selecting regex input
		$regexInput = isset($_GET['regexInput']); // Custom regex input

		if (strlen($q) == 0) {
			return '<div class="search">'.$this->getSearchForm($webdir).'</div>';
		}

		// Nobody likes suppressed errors with @ but I'm doing it anyway here
		if ($regexInput && @preg_match($q, '') === false) {
			$err = error_get_last()['message'] ?? '';
			error_clear_last(); // Clear error
			return '<div class="search">'.$this->getSearchForm($webdir, $q) . '<p class="red">Bad regex input: <strong>'.htmlspecialchars($q).'</strong><br />'.$err.'</p></div>';
		}

		$time = -microtime(true); // Measure execution time

		// Read all files down through $dir
		if (!$this->readDirContent($dir)) {
			return '<div class="search">'.$this->getSearchForm($webdir, $q) . '<p class="red">Could not read current directory.</p></div>';
		}

		// Do the search
		$search_res = $this->search($this->allFiles, $q, $wholeWord, $regexInput);
		$results = '';
	
		if (count($search_res['previews']) == 0) {
			$results = '<div class="searchResult red"><strong>No matches found.</strong></div>';
		}
		else {
			$results = implode('', $search_res['previews']);
		}

		$time += microtime(true);

		return '<div class="search">'.$this->getSearchForm($webdir, $q).
			'<div class="searchResultContainer">'.
			'<h3 class="searchResultsHeader">'.$search_res['header'].' ('.round($time, 3).' s)</h3>'.
			$results.
			$search_res['stats'].
			'</div></div>';
	}

	// ====================================================================
	// Search form.
	// --------------------------------------------------------------------
	private function getSearchForm(string $webdir, string $q = ''): string
	{
		return
			'<form action="'.htmlspecialchars($_SERVER['PHP_SELF']).'" method="get">
			<fieldset>
			<legend>Search Site</legend>
			<input type="search" id="q" value="'.htmlspecialchars($q).'" onkeypress="fmFindEnter(event)" autofocus="autofocus" />
			<button type="button" onclick="fmFind()">Search</button>
			<div class="miniAdd">
				Powered by 
				<a href="https://www.dunweber.com/docs/scripts/releases/searchengine"
					title="Get your own search engine from dunweber.com!">
					<strong>dunweber.com</strong>
				</a>
			</div>
			<label><input type="checkbox" id="wholeWord" value="1"'.(isset($_GET['wholeWord']) ? ' checked="checked"' : '').' /> Whole Word</label> 
			<label><input type="checkbox" id="regexInput" value="1"'.(isset($_GET['regexInput']) ? ' checked="checked"' : '').' /> Regex</label>
			</fieldset>
			</form>
			<script>
			function fmFindEnter(event)
			{
				if (event.key === "Enter") {
					event.preventDefault();
					fmFind();
				}
			}
			function fmFind()
			{
				var str = document.getElementById(\'q\').value;
				var whole = document.getElementById(\'wholeWord\').checked ? "&wholeWord=1" : "";
				var regex = document.getElementById(\'regexInput\').checked ? "&regexInput=1" : "";
				if (str.length > 0) {
					window.location = "?q="+encodeURIComponent(str)+whole+regex;
				}
			}
			</script>';
	}

	// ====================================================================
	// Read all files in all subdirectories and store in global array.
	// --------------------------------------------------------------------
	private function readDirContent(string $dir): bool
	{
		if (($content = scandir($dir, SCANDIR_SORT_NONE)) === false) {
			return false;
		}
	
		$content = array_diff($content, ['.','..']); // Skip this and parent dir
	
		foreach ($content as $name) {
			$file = $dir.DIRECTORY_SEPARATOR.$name;
	
			if (is_link($file)){ // Skip symbolic links
				continue;
			}
			if (is_dir($file)) {
				$this->readDirContent($file);
			}
			if (is_file($file)) {
				$this->allFiles[] = $file;
			}
		}
	
		return true;
	}

	// ====================================================================
	// Check if file type should be opened and content searched through.
	// --------------------------------------------------------------------
	private function isContentReadable(string $path): bool
	{
		$ext = strtolower(pathinfo($path, PATHINFO_EXTENSION));
	
		for ($i = 0; $i < count(Conf::SEARCH_SOURCE_TYPES); $i++) {
			if ($ext === Conf::SEARCH_SOURCE_TYPES[$i]) {
				return true;
			}
		}

		return false;
	}

	// ====================================================================
	// Search for matches to $q in all files in the $files array.
	// --------------------------------------------------------------------
	private function search(array $files, string $q, bool $wholeWord, bool $regexInput): array
	{
		$num_files = count($files);
		$pages_with_searchable_content = 0;
		$size_total_searched = 0;
		$skipped_as_too_big = 0;
		$max_match = 0;
		$previews = [];
		$hits = [];

		for ($i = 0; $i < $num_files; $i++) {
			$path = $files[$i];
			$content = '';
			$title = '';
			$filename = basename($path);

			if ($this->isContentReadable($path)) {
				$tooBig = filesize($path) > Conf::FILESIZE_MAX_SEARCH_OPEN;

				if (!$tooBig) {
					$content = file_get_contents($path) ?? '';

					// Extract title before stripping tags
					$title = $this->getTitle($content, $path);

					// Strip tags as not relevant for normal users to search in tags and code
					$content = strip_tags($content);
	
					if (strlen($content) > 0) {
						$pages_with_searchable_content++;
					}
				}
				else {
					$skipped_as_too_big++;
				}
			}
			else {
				$title = $filename;
			}

			// Count bytes searched for stats
			// Well, counts filename twice when no title was found and counts title twice if title came from within content (<title>), can I live with it?
			$size_total_searched += strlen($content) + strlen($title) + strlen($filename);

			// Check for matches to $q
			$match = $this->match($q, $wholeWord, $regexInput, $content, $title, $path);

			if (count($match) > 0) {
				$previews[] = $match['preview'];
				$hits[] = $match['hits'];
			}
		}

		$num_pages_with_matches = count($previews);

		if ($num_pages_with_matches > 0) {
			// Sort so most hits are shown first
			array_multisort($hits, SORT_DESC, SORT_NUMERIC, $previews);
			$max_match = $hits[0];
		}

		$search_res['header'] = '<span class="searchResTitle">Search results for &quot;'.htmlspecialchars($q).'&quot;</span>
			<span class="searchResSubTitle">Total '.(int)$num_pages_with_matches.'</span>';

		$search_res['previews'] = $previews;

		$search_res['stats'] =
			'<hr />
			<table class="searchStats">
			<tr><th class="statsTitle" colspan="2"><strong>Statistics:</strong></th></tr>
			<tr><th>Searching in files:</th><td>'.(int)$num_files.'</td></tr>
			<tr><th>Files with readable content:</th><td>'.(int)$pages_with_searchable_content.'</td></tr>
			<tr><th>Readable files skipped as too big:</th><td>'.(int)$skipped_as_too_big.'</td></tr>
			<tr><th>Size of searched content:</th><td>'.$this->getFriendlySize($size_total_searched, 0).'</td></tr>
			<tr><th>Files with matches:</th><td>'.(int)$num_pages_with_matches.'</td></tr>
			<tr><th>File with most matches has:</th><td>'.(int)$max_match.'</td></tr>
			</table>
			<hr />';

		return $search_res;
	}

	// ====================================================================
	// Search for matches to $q in $content and in the basename of $path
	// using preg_match_all().
	// --------------------------------------------------------------------
	private function match(string $q, bool $wholeWord, bool $regexInput, string $content, string $title, string $path): array
	{
		// Sanitize query by escaping delimiter (only applicable for non-regex input)
		$q_san = preg_quote($q, '/');

		// Do the search / check for matches
		$pattern = $regexInput ? $q : ($wholeWord ? '/\b('.$q_san.')\b/iu' : '/[^a-zA-Z]*('.$q_san.')[^a-zA-Z]*/iu');

		$hits_content = preg_match_all($pattern, $content, $matches, PREG_OFFSET_CAPTURE);
		$hits_title = preg_match_all($pattern, $title); // Must search title separately as my titles are stored in a PHP var and hence stripped and not giving hits when searching content in line above
		$hits_filename = preg_match_all($pattern, basename($path)); // Will be same hit as line above if no title was found (because then title == filename)

		$hits = $hits_content + $hits_title + $hits_filename;

		if ($hits == 0) {
			return []; // No matches
		}

		// Building preview of results
		$preview = $this->makePreview($q, $wholeWord, $regexInput, $path, $content, $title, $matches, $hits_content, $hits_filename, $hits_title);

		return ['preview' => $preview, 'hits' => $hits];
	}

	// ====================================================================
	// Make preview of matching search result.
	// --------------------------------------------------------------------
	private function makePreview(string $q, bool $wholeWord, bool $regexInput, string $path, string $content, string $title, array $matches, int $hits_content, int $hits_filename, int $hits_title): string
	{
		$preview_content_length = 120;
		$preview_title_length = 60;
		$offset = $matches[0][0][1] ?? 0;

		// Remove line breaks for preview
		$preview_content = trim(preg_replace('/\n|\r|\t/iu',' ',$content));
		$preview_title = $title;

		// Make sure offset is inside preview
		if ($offset + $preview_content_length > strlen($preview_content)) {
			$offset = strlen($preview_content) - $preview_content_length;
		}
		if ($offset < $preview_content_length) {
			$offset = 0;
		}

		// Matches are highligthed in bold in the previews
		$pattern_highlight = '/([.]*)('.preg_quote(htmlspecialchars($regexInput ? ($matches[0][0][0] ?? '') : $q), '/').')([.]*)/iu';
		$pattern_replace = '\\1<strong>\\2</strong>\\3';

		// Don't hightlight partial matches when whole word is selected
		if ($wholeWord && !$regexInput) {
			$pattern_highlight = '/([.]*)\b('.preg_quote(htmlspecialchars($q), '/').')\b([.]*)/iu';
		}

		// Make preview of content
		if (strlen($preview_content) > $preview_content_length) {
			$preview_content = substr($preview_content, $offset, $preview_content_length).' ...';
		}
		if ($hits_content > 0) {
			$preview_content = preg_replace($pattern_highlight, $pattern_replace, htmlspecialchars($preview_content));
		}
		else {
			$preview_content = '[Match on title or filename]';
		}
		if (strlen($preview_content) == 0) {
			$preview_content = '<em>[no preview]</em>';
		}

		// Make preview of title (which is filename in case no title was found)
		if (strlen($preview_title) > $preview_title_length) {
			$preview_title = substr($preview_title, 0, $preview_title_length).' ...';
		}
		if ($hits_title > 0 || $hits_filename > 0) {
			$preview_title = preg_replace($pattern_highlight, $pattern_replace, htmlspecialchars($preview_title));
		}

		// Title matches count as content
		if ($hits_title > 0) {
			$hits_content += $hits_title;
		}

		// Create link to the file with the hit
		$link = htmlspecialchars($this->getWebPath($path));

		return
			'<div class="searchResult"><a href="'.$link.'">'.$preview_title.'</a>
			<span class="searchResMiniStat">[content: '.(int)$hits_content.', filename: '.(int)$hits_filename.']</span><br />
			<span class="searchResContentPre">'.$preview_content.'</span><br />
			<span class="searchResPathPre">'.$link.'</span>
			</div>';
	}

	// ====================================================================
	// Extract title for showing as link text if possible else use filename.
	// --------------------------------------------------------------------
	private function getTitle(string $content, string $path): string
	{
		// Use something like this if your site's titles are stored in a PHP var
		//$title = $this->extractTag($content, '$title = \'', '\'');
		$title = '';

		if (strlen($title) == 0) {
			$title = $this->extractTag($content, '<title>', '</title>');
		}

		if (strlen($title) == 0) {
			$title = basename($path);
		}
		
		return $title;
	}

	// ====================================================================
	// Extracts the content between to strings, e.g. content of a HTML tag.
	// E.g. <title>Hello</title> gives the string "Hello".
	// --------------------------------------------------------------------
	private function extractTag(string $str, string $startTag, string $endTag): string
	{ 
		preg_match('~'.preg_quote($startTag, '~').'(.*?)'.preg_quote($endTag, '~').'~iu', $str, $matches); 
		return $matches[1] ?? ''; 
	} 

	// ====================================================================
	// Get web path from real path.
	// --------------------------------------------------------------------
	private function getWebPath(string $path): string
	{
		return str_replace('\\', '/', mb_substr($path, mb_strlen(realpath($_SERVER['DOCUMENT_ROOT']))));
	}

	// ====================================================================
	// Present data size in bytes as human-readable with SI prefix.
	// Parameter $bytesize will normally come from filesize().
	// --------------------------------------------------------------------
	private function getFriendlySize(int $bytesize, int $precision = 2): string
	{
		$sizes = array('YB', 'ZB', 'EB', 'PB', 'TB', 'GB', 'MB', 'kB', 'B');
		$total = count($sizes);
	
		while ($total-- && $bytesize > 1024) {$bytesize /= 1024;}
		return round($bytesize, $precision).' '.$sizes[$total];
	}
}
?>
